@@ -1,3 +1,7 @@
+# Copyright (c) Jupyter Development Team.
+# Distributed under the terms of the Modified BSD License.
+#
+
 import copy
 
 import json
@@ -34,21 +38,24 @@ allowed_cursor = ['alias', 'cell', 'grab', 'move', 'crosshair', 'context-menu',
                   'progress', 'text', 'wait', 'zoom-in', 'zoom-out']
 
 
-def basemap_to_tiles(bm, day='yesterday', **kwargs):
+def basemap_to_tiles(basemap, day='yesterday', **kwargs):
     # Format the URL with modisdate
     from datetime import date, timedelta
+
     if day == 'yesterday':
         yesterday = date.today() - timedelta(1)
         day = yesterday.strftime('%Y-%m-%d')
-    url = bm.get('url', '')
+
+    url = basemap.get('url', '')
     if url.count('%'):
         url = url % day
+
     return TileLayer(
         url=url,
-        max_zoom=bm.get('max_zoom', 19),
-        min_zoom=bm.get('min_zoom', 1),
-        attribution=bm.get('attribution', ''),
-        name=bm.get('name', ''),
+        max_zoom=basemap.get('max_zoom', 19),
+        min_zoom=basemap.get('min_zoom', 1),
+        attribution=basemap.get('attribution', ''),
+        name=basemap.get('name', ''),
         **kwargs
     )
 
@@ -89,7 +96,7 @@ class Layer(Widget, InteractMixin):
     popup_max_width = Int(300).tag(sync=True)
     popup_max_height = Int(default_value=None, allow_none=True).tag(sync=True)
 
-    options = List(trait=Unicode).tag(sync=True)
+    options = List(trait=Unicode()).tag(sync=True)
 
     def __init__(self, **kwargs):
         super(Layer, self).__init__(**kwargs)
@@ -159,6 +166,19 @@ class Icon(UILayer):
     popup_anchor = Tuple((0, 0), allow_none=True).tag(sync=True, o=True)
 
 
+class AwesomeIcon(UILayer):
+    _view_name = Unicode('LeafletAwesomeIconView').tag(sync=True)
+    _model_name = Unicode('LeafletAwesomeIconModel').tag(sync=True)
+
+    name = Unicode('home').tag(sync=True)
+    marker_color = Enum(
+        values=['white', 'red', 'darkred', 'lightred', 'orange', 'beige', 'green', 'darkgreen', 'lightgreen', 'blue', 'darkblue', 'lightblue', 'purple', 'darkpurple', 'pink', 'cadetblue', 'white', 'gray', 'lightgray', 'black'],
+        default_value='blue'
+    ).tag(sync=True)
+    icon_color = Color('white').tag(sync=True)
+    spin = Bool(False).tag(sync=True)
+
+
 class Marker(UILayer):
     _view_name = Unicode('LeafletMarkerView').tag(sync=True)
     _model_name = Unicode('LeafletMarkerModel').tag(sync=True)
@@ -166,7 +186,7 @@ class Marker(UILayer):
     location = List(def_loc).tag(sync=True)
     opacity = Float(1.0, min=0.0, max=1.0).tag(sync=True)
     visible = Bool(True).tag(sync=True)
-    icon = Instance(Icon, allow_none=True, default_value=None).tag(sync=True, **widget_serialization)
+    icon = Union((Instance(Icon), Instance(AwesomeIcon)), allow_none=True, default_value=None).tag(sync=True, **widget_serialization)
 
     # Options
     z_index_offset = Int(0).tag(sync=True, o=True)
@@ -239,6 +259,9 @@ class TileLayer(RasterLayer):
     attribution = Unicode('Map data (c) <a href="https://openstreetmap.org">OpenStreetMap</a> contributors').tag(sync=True, o=True)
     detect_retina = Bool(False).tag(sync=True, o=True)
     no_wrap = Bool(False).tag(sync=True, o=True)
+    tms = Bool(False).tag(sync=True, o=True)
+    show_loading = Bool(False).tag(sync=True)
+    loading = Bool(False, read_only=True).tag(sync=True)
 
     _load_callbacks = Instance(CallbackDispatcher, ())
 
@@ -253,7 +276,7 @@ class TileLayer(RasterLayer):
     def on_load(self, callback, remove=False):
         self._load_callbacks.register_callback(callback, remove=remove)
 
-    def redraw(self): 
+    def redraw(self):
         self.send({'msg':'redraw'})
 
 class LocalTileLayer(TileLayer):
@@ -312,7 +335,7 @@ class Velocity(Layer):
     longitude_dimension = Unicode('longitude', help='Name of the longitude dimension in the dataset')
     units = Unicode(None, allow_none=True)
 
-    data = Dataset().tag(sync=True, to_json=ds_x_to_json)
+    data = Dataset().tag(dtype=None, sync=True, to_json=ds_x_to_json)
 
     # Options
     display_values = Bool(True).tag(sync=True, o=True)
@@ -376,7 +399,7 @@ class Path(VectorLayer):
     color = Color('#0033FF').tag(sync=True, o=True)
     weight = Int(5).tag(sync=True, o=True)
     fill = Bool(True).tag(sync=True, o=True)
-    fill_color = Color('#0033FF').tag(sync=True, o=True)
+    fill_color = Color(None, allow_none=True).tag(sync=True, o=True)
     fill_opacity = Float(0.2).tag(sync=True, o=True)
     dash_array = Unicode(allow_none=True, default_value=None).tag(sync=True, o=True)
     line_cap = Unicode('round').tag(sync=True, o=True)
@@ -386,15 +409,39 @@ class Path(VectorLayer):
     opacity = Float(1.0, min=0.0, max=1.0).tag(sync=True, o=True)
 
 
+class AntPath(VectorLayer):
+    _view_name = Unicode('LeafletAntPathView').tag(sync=True)
+    _model_name = Unicode('LeafletAntPathModel').tag(sync=True)
+
+    locations = List().tag(sync=True)
+
+    # Options
+    use = Enum(values=['polyline', 'polygon', 'rectangle', 'circle'], default_value='polyline').tag(sync=True, o=True)
+    delay = Int(400).tag(sync=True, o=True)
+    weight = Int(5).tag(sync=True, o=True)
+    dash_array = List([10, 20]).tag(sync=True, o=True)
+    color = Color('#0000FF').tag(sync=True, o=True)
+    pulse_color = Color('#FFFFFF').tag(sync=True, o=True)
+    paused = Bool(False).tag(sync=True, o=True)
+    reverse = Bool(False).tag(sync=True, o=True)
+    hardware_accelerated = Bool(False).tag(sync=True, o=True)
+    radius = Int(10).tag(sync=True, o=True)
+
+
 class Polyline(Path):
     _view_name = Unicode('LeafletPolylineView').tag(sync=True)
     _model_name = Unicode('LeafletPolylineModel').tag(sync=True)
 
     locations = List().tag(sync=True)
+    scaling = Bool(True).tag(sync=True)
+    rotation = Bool(True).tag(sync=True)
+    uniform_scaling = Bool(False).tag(sync=True)
 
     # Options
     smooth_factor = Float(1.0).tag(sync=True, o=True)
     no_clip = Bool(True).tag(sync=True, o=True)
+    transform = Bool(False).tag(sync=True, o=True)
+    draggable = Bool(False).tag(sync=True, o=True)
 
 
 class Polygon(Polyline):
@@ -431,14 +478,14 @@ class MarkerCluster(Layer):
     _view_name = Unicode('LeafletMarkerClusterView').tag(sync=True)
     _model_name = Unicode('LeafletMarkerClusterModel').tag(sync=True)
 
-    markers = Tuple(trait=Instance(Marker)).tag(sync=True, **widget_serialization)
+    markers = Tuple().tag(trait=Instance(Marker), sync=True, **widget_serialization)
 
 
 class LayerGroup(Layer):
     _view_name = Unicode('LeafletLayerGroupView').tag(sync=True)
     _model_name = Unicode('LeafletLayerGroupModel').tag(sync=True)
 
-    layers = Tuple(trait=Instance(Layer)).tag(sync=True, **widget_serialization)
+    layers = Tuple().tag(trait=Instance(Layer), sync=True, **widget_serialization)
 
     _layer_ids = List()
 
@@ -466,6 +513,13 @@ class LayerGroup(Layer):
             raise LayerException('layer not on in layergroup: %r' % layer)
         self.layers = tuple([l for l in self.layers if l.model_id != layer.model_id])
 
+    def substitute_layer(self, old, new):
+        if isinstance(new, dict):
+            new = basemap_to_tiles(new)
+        if old.model_id not in self._layer_ids:
+            raise LayerException('Could not substitute layer: layer not in layergroup.')
+        self.layers = tuple([new if l.model_id == old.model_id else l for l in self.layers])
+
     def clear_layers(self):
         self.layers = ()
 
@@ -482,6 +536,7 @@ class GeoJSON(FeatureGroup):
     data = Dict().tag(sync=True)
     style = Dict().tag(sync=True)
     hover_style = Dict().tag(sync=True)
+    point_style = Dict().tag(sync=True)
 
     _click_callbacks = Instance(CallbackDispatcher, ())
     _hover_callbacks = Instance(CallbackDispatcher, ())
@@ -514,7 +569,7 @@ class GeoData(GeoJSON):
     geo_dataframe = Instance('geopandas.GeoDataFrame')
 
     def __init__(self, **kwargs):
-        super(GeoJSON, self).__init__(**kwargs)
+        super(GeoData, self).__init__(**kwargs)
         self.data = self._get_data()
 
     @observe('geo_dataframe')
@@ -584,7 +639,7 @@ class Control(Widget):
     _view_module_version = Unicode(EXTENSION_VERSION).tag(sync=True)
     _model_module_version = Unicode(EXTENSION_VERSION).tag(sync=True)
 
-    options = List(trait=Unicode).tag(sync=True)
+    options = List(trait=Unicode()).tag(sync=True)
 
     position = Enum(
         ['topright', 'topleft', 'bottomright', 'bottomleft'],
@@ -788,6 +843,16 @@ class ZoomControl(Control):
     zoom_out_title = Unicode('Zoom out').tag(sync=True, o=True)
 
 
+class ScaleControl(Control):
+    _view_name = Unicode('LeafletScaleControlView').tag(sync=True)
+    _model_name = Unicode('LeafletScaleControlModel').tag(sync=True)
+
+    max_width = Int(100).tag(sync=True, o=True)
+    metric = Bool(True).tag(sync=True, o=True)
+    imperial = Bool(True).tag(sync=True, o=True)
+    update_when_idle = Bool(False).tag(sync=True, o=True)
+
+
 class AttributionControl(Control):
     _view_name = Unicode('LeafletAttributionControlView').tag(sync=True)
     _model_name = Unicode('LeafletAttributionControlModel').tag(sync=True)
@@ -824,11 +889,13 @@ class Map(DOMWidget, InteractMixin):
     crs = Enum(values=allowed_crs, default_value='EPSG3857').tag(sync=True)
 
     # Specification of the basemap
-    basemap = Dict(default_value=dict(
+    basemap = Union(
+        (Dict(), Instance(TileLayer)),
+        default_value=dict(
             url='https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png',
             max_zoom=19,
             attribution='Map data (c) <a href="https://openstreetmap.org">OpenStreetMap</a> contributors'
-        )).tag(sync=True, o=True)
+        ))
     modisdate = Unicode('yesterday').tag(sync=True)
 
     # Interaction options
@@ -855,17 +922,14 @@ class Map(DOMWidget, InteractMixin):
     # marker_zoom_animation = Bool(?).tag(sync=True, o=True)
     fullscreen = Bool(False).tag(sync=True, o=True)
 
-    options = List(trait=Unicode).tag(sync=True)
+    options = List(trait=Unicode()).tag(sync=True)
 
     style = InstanceDict(MapStyle).tag(sync=True, **widget_serialization)
     default_style = InstanceDict(MapStyle).tag(sync=True, **widget_serialization)
     dragging_style = InstanceDict(MapStyle).tag(sync=True, **widget_serialization)
-    
-    zoom_control = Bool(True)
-    zoom_control_instance = ZoomControl()
 
+    zoom_control = Bool(True)
     attribution_control = Bool(True)
-    attribution_control_instance = AttributionControl(position='bottomright')
 
     @default('dragging_style')
     def _default_dragging_style(self):
@@ -880,11 +944,15 @@ class Map(DOMWidget, InteractMixin):
     east = Float(def_loc[1], read_only=True).tag(sync=True)
     west = Float(def_loc[1], read_only=True).tag(sync=True)
 
-    layers = Tuple(trait=Instance(Layer)).tag(sync=True, **widget_serialization)
+    layers = Tuple().tag(trait=Instance(Layer), sync=True, **widget_serialization)
 
     @default('layers')
     def _default_layers(self):
-        return (basemap_to_tiles(self.basemap, self.modisdate, base=True),)
+        basemap = self.basemap if isinstance(self.basemap, TileLayer) else basemap_to_tiles(self.basemap, self.modisdate)
+
+        basemap.base = True
+
+        return (basemap, )
 
     bounds = Tuple(read_only=True)
     bounds_polygon = Tuple(read_only=True)
@@ -899,38 +967,38 @@ class Map(DOMWidget, InteractMixin):
                                           (self.south, self.west)))
 
     def __init__(self, **kwargs):
+        self.zoom_control_instance = None
+        self.attribution_control_instance = None
+
         super(Map, self).__init__(**kwargs)
-        self.on_displayed(self._fire_children_displayed)
         self.on_msg(self._handle_leaflet_event)
 
         if self.zoom_control:
+            self.zoom_control_instance = ZoomControl()
             self.add_control(self.zoom_control_instance)
-            
+
         if self.attribution_control:
+            self.attribution_control_instance = AttributionControl(position='bottomright')
             self.add_control(self.attribution_control_instance)
-        
+
     @observe('zoom_control')
     def observe_zoom_control(self, change):
         if change['new']:
+            self.zoom_control_instance = ZoomControl()
             self.add_control(self.zoom_control_instance)
         else:
-            if self.zoom_control_instance in self.controls:
+            if self.zoom_control_instance is not None and self.zoom_control_instance in self.controls:
                 self.remove_control(self.zoom_control_instance)
 
     @observe('attribution_control')
     def observe_attribution_control(self, change):
         if change['new']:
+            self.attribution_control_instance = AttributionControl(position='bottomright')
             self.add_control(self.attribution_control_instance)
         else:
-            if self.attribution_control_instance in self.controls:
+            if self.attribution_control_instance is not None and self.attribution_control_instance in self.controls:
                 self.remove_control(self.attribution_control_instance)
 
-    def _fire_children_displayed(self, widget, **kwargs):
-        for layer in self.layers:
-            layer._handle_displayed(**kwargs)
-        for control in self.controls:
-            control._handle_displayed(**kwargs)
-
     _layer_ids = List()
 
     @validate('layers')
@@ -967,7 +1035,7 @@ class Map(DOMWidget, InteractMixin):
     def clear_layers(self):
         self.layers = ()
 
-    controls = Tuple(trait=Instance(Control)).tag(sync=True, **widget_serialization)
+    controls = Tuple().tag(trait=Instance(Control), sync=True, **widget_serialization)
     _control_ids = List()
 
     @validate('controls')
